/*******************************************************************************\
 * Copyright (c) 1997 by Markus Koelblin
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
\*******************************************************************************/

/******************************************************************************\
 * mclock.c
 *
 * Simple digital clock for the X-Window system including
 * mail checker and average system load monitor.
 *
 * Author:  Markus Koelblin (Markus.Koelblin@hadiko.de) 1997
 *
\******************************************************************************/

#include <pwd.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <sys/stat.h>
#include "mclock.h"

#include "mclock.xbm"
#include "flagup.xbm"
#include "flagdn.xbm"


/******************************************************************************\
 * Type definitions
\******************************************************************************/

typedef struct tag_MailStruct
{
    XPoint   position;        /* Position of the mailbox icon    */
    char     *filename;       /* The mailbox file to be watched  */
    char     *check_cmd;      /* Optional command for mail check */
    char     *sound_cmd;      /* Sound command on new mail click */
    long     lastsize;	      /* Size in bytes of mailbox file   */
    Boolean  flagup;	      /* Current state of mailbox icon   */
    Boolean  redraw;
    Pixmap   icon_up;         /* Icons used for mailbox display  */
    Pixmap   icon_dn;
}
MailStruct;

typedef struct tag_ClockStruct
{
    XFontStruct	*font;	      /* The font used for text display */
    XRectangle  rect;         /* Clock area in the main window  */
    GC          gc;           /* Graphics context for output    */
    char        text[8];      /* Time string to be displayed    */
    int         minute, hour; /* Current time values            */
}
ClockStruct;

typedef struct tag_LoadStruct
{
    XRectangle  rect;         /* Histogram area in the main window */
    Pixmap      chart;        /* Pixmap for histogram display      */
    double      load;         /* Actual average system load        */
    short       xpos;         /* Current position in histgram      */
    short       xofs;         /* Offset used for left scrolling    */
}
LoadStruct;

typedef struct tag_ClockApp
{
    ClockStruct   clock;
    MailStruct    mail;
    LoadStruct    load;
    Pixmap        icon;
    GC            gc;
}
ClockAppStruct;


/******************************************************************************\
 * Constants
\******************************************************************************/

#define EV_MASK     ExposureMask | ButtonPressMask | KeyPressMask
#define CW_MASK     CWBackPixel  | CWBorderPixel | CWBackPixmap | CWBorderPixmap | \
CWBitGravity | CWWinGravity  | CWColormap   | CWBackingStore

#define PADDING     4
#define DEF_UPDATE  30 
#define WINDOW_POS  0
#define WINDOW_SIZE 56

/* Drawing flags for app_window_draw */
#define MF_CLOCK (1 << 0)
#define MF_MAIL  (1 << 1)
#define MF_LOAD  (1 << 2)
#define MF_CLEAR (1 << 3)
#define MF_ALL   (MF_CLOCK | MF_MAIL | MF_LOAD | MF_CLEAR)

/*
 * Error message constants
 */
enum { ERR_DISPLAY=0, ERR_COLOR, ERR_USERNAME, ERR_SIGNAL, ERR_MALLOC, ERR_COUNT }; 

const char error_messages[ERR_COUNT][32] =
{
    "can't connect to display",
    "can't allocate color",
    "can't get username",
    "can't setup signal handler",
    "memory allocation failed",
};

/*
 * Help text on wrong options
 */
const char usage_text[] = "Usage:\n\
  mclock [-display <name>] [-geometry <geom>] [-fg <color>] [-bg <color>]\n\
         [-fn <font>] [-update <sec>] [-wav <sound>] [-exec <cmd>]\n";


/******************************************************************************\
 * Globals
\******************************************************************************/

static char font_default[] = "-adobe-times-*-*-*-*-*-180-*";
static char back_default[] = "grey75";
static char class_name[]   = "Mclock";

static XSetWindowAttributes attributes;
static Atom     wm_delete_win, wm_protocols;
static Display  *display;
static Screen   *screen;
static Window   main_win;
static Pixel    black_pix, white_pix;
static Pixel    fg_pix, bg_pix;

/*
 * Option values
 */
static char  *display_name = NULL;
static char  *geom_string  = NULL;
static char  *file_name = NULL;
static char  *exec_name = NULL;
static char  *wav_name = NULL;
static char  *fn_name = NULL;
static char  *bg_name = NULL;
static char  *fg_name = NULL;
static int   update = DEF_UPDATE;

/*
 * Application data
 */
static ClockAppStruct app_data;


/******************************************************************************\
 * Functional macros and Function prototypes
\******************************************************************************/

#define ERROR_MSG(nr, str) \
 fprintf(stderr, "%s: %s %s\n", class_name, error_messages[nr], (str));

#define SET_RECT(r, xr, yr, wr, hr) \
 (r).x      = (xr); \
 (r).y      = (yr); \
 (r).width  = (wr); \
 (r).height = (hr);

void clock_setup(ClockStruct* clk);
void clock_alarm(int sig_no);

void mail_setup(MailStruct* mailbox);
void mail_check(MailStruct* mailbox, Boolean reset);

void app_window_paint(ClockAppStruct* app, unsigned mode);
void app_window_setup(ClockAppStruct* app);
void app_window_close(int code);

void draw_border(const XRectangle* rect, GC gc, Pixel low, Pixel high);
int  get_options(int argc, char* argv[]);


/******************************************************************************\
 * alloc_color()
\******************************************************************************/

static Pixel alloc_color(Display* display, String name, Pixel fallback_pix)
{
    if (name)
    {
	Colormap cmap = DefaultColormap(display, DefaultScreen(display));
	XColor   col;

	if (XParseColor(display, cmap, name, &col) && XAllocColor(display, cmap, &col))
	    return col.pixel;
	else
	    ERROR_MSG(ERR_COLOR, name);
    }
    return fallback_pix;
}

/******************************************************************************\
 * save_malloc()
\******************************************************************************/

static void* save_malloc(size_t n_bytes)
{
    void* block_ptr = malloc(n_bytes);

    if (block_ptr == NULL)
    {
	ERROR_MSG(ERR_MALLOC, "(save_malloc)");
	exit(-1);
    }
    return block_ptr;
}

/******************************************************************************\
 * main()
\******************************************************************************/

void main(int argc, char* argv[])
{
    XSizeHints *size_hints;
    XClassHint *class_hint;
    XWMHints   *wm_hints; 
    int        x, y, width, height;
    int        screen_nr;
    int        depth;

    if (!get_options(argc, argv))
	exit(-1);

    /* Open a connection to the Display */
    if ((display = XOpenDisplay(display_name)) == NULL)
    {
	ERROR_MSG(ERR_DISPLAY, "");
        exit(-1);
    }

    /* Get some default values of the server */
    screen    = DefaultScreenOfDisplay(display);
    screen_nr = DefaultScreen(display); 
    depth     = DefaultDepthOfScreen(screen);
    black_pix = BlackPixelOfScreen(screen);
    white_pix = WhitePixelOfScreen(screen);

    /* Allocate foreground and background colors */
    fg_pix = alloc_color(display, fg_name, black_pix);
    bg_pix = alloc_color(display, bg_name, white_pix);

    /* Setting the window attributes */
    attributes.background_pixel  = bg_pix;
    attributes.background_pixmap = ParentRelative;
    attributes.border_pixel      = None;
    attributes.backing_store     = NotUseful;
    attributes.bit_gravity       = CenterGravity;
    attributes.win_gravity       = StaticGravity;
    attributes.colormap          = DefaultColormap(display, screen_nr);

    /* Get initial window size and position */
    if (geom_string == NULL || !XParseGeometry(geom_string, &x, &y, &width, &height))
    {
	x = y = WINDOW_POS;
	width = height = WINDOW_SIZE;
    }

    /* Creating the top-level window with the given attributes */
    main_win = XCreateWindow(display,
			     DefaultRootWindow(display),
			     x, y, width, height, 0,
			     DefaultDepthOfScreen(screen),
			     InputOutput,
			     DefaultVisualOfScreen(screen),
			     CW_MASK,
			     &attributes);

    /* Create application icon for mclock */
    app_data.icon = XCreatePixmapFromBitmapData(display, main_win, (char*) mclock_bits,
						mclock_width, mclock_height,
						fg_pix, bg_pix, depth); 
    /*
     * Hints for the Windowmanager
     */
    wm_hints = XAllocWMHints();
    wm_hints->icon_x      = 0;
    wm_hints->icon_y      = 0;
    wm_hints->input       = True;
    wm_hints->icon_pixmap = app_data.icon;
    wm_hints->flags       = InputHint | IconPixmapHint;
    XSetWMHints(display, main_win, wm_hints);

    /* Set size hints and standard properties */
    size_hints = XAllocSizeHints();
    size_hints->min_width  = 32;
    size_hints->min_height = 32;
    size_hints->max_width  = 256;
    size_hints->max_height = 256;
    size_hints->flags      = PMinSize | PMaxSize;
    XSetStandardProperties(display, main_win, "mclock", NULL, app_data.icon,
			   argv, argc, size_hints);

    /* Set class and size hints */
    class_hint = XAllocClassHint();
    class_hint->res_name  = class_name;
    class_hint->res_class = class_name;
    XSetClassHint(display, main_win, class_hint);

    XFree(wm_hints);
    XFree(size_hints);
    XFree(class_hint);

    /* Make sure that we get the WM_DELETE_WINDOW message */
    wm_protocols  = XInternAtom(display, "WM_PROTOCOLS", False);
    wm_delete_win = XInternAtom(display, "WM_DELETE_WINDOW", False);
    XSetWMProtocols(display, main_win, &wm_delete_win, 1);

    /* Setting the event mask for window and map (show) all windows*/
    XSelectInput(display, main_win, EV_MASK | StructureNotifyMask);
    XMapWindow(display, main_win);

    /* Initialize and setup the alarm handler */
    app_window_setup(&app_data);

    /* Register and activate clock alarm funtion */
    if (signal(SIGALRM, clock_alarm) == SIG_ERR)
    {
	ERROR_MSG(ERR_SIGNAL, "(SIGALRM)");
	app_window_close(-1);
    }
    alarm(update);

    /***********************************************************\
     * Mclock's event loop
     *
     * Force geometry calculation in first ConfigurNotify event
    \***********************************************************/
    width = height = 0;

    while (1)
    {
	XClientMessageEvent *cm_ptr;
	XEvent              event;
	KeySym              keysym;

        XNextEvent(display, &event);

        switch (event.type)
	{
	case Expose:

            /* Draw window labels when becomming visible */
	    if (event.xany.window == main_win)
	    {
		app_data.mail.redraw = True;
		app_window_paint(&app_data, MF_ALL);
	    }
	    break;


	case ButtonPress:

	    /* Catch the button press events */
	    if (event.xany.window == main_win)
	    {
		if (exec_name) system(exec_name);
		mail_check(&app_data.mail, True);
		app_window_paint(&app_data, MF_MAIL);
	    }
	    break;


	case KeyPress:

	    /* Lookup the key symbol */
	    keysym = XLookupKeysym(&event.xkey, 1);

	    if (keysym == XK_Q || keysym == XK_q)
		XDestroyWindow(display, main_win);
	    break;


	case ConfigureNotify:

	    /* Catch the ConfigurNotify event */
	    if (event.xany.window == main_win)
	    {
		XRectangle *rect;
		unsigned   w, h, bw;
		Window     root;

		XGetGeometry(display, main_win, &root, &x, &y, &w, &h, &bw, &depth);

		if (w == width || h == height)
		    break;

		/* Else resize mclock's window */
		width  = w;
		height = h;
		h = h / 2;

		app_data.mail.position.x   = PADDING;
		app_data.mail.position.y   = (h - flagup_height) / 2;
		app_data.clock.rect.x      = PADDING;
		app_data.clock.rect.y      = h + PADDING;
		app_data.clock.rect.width  = w - PADDING*2;
		app_data.clock.rect.height = h - PADDING*2;
		app_data.load.xpos         = 0;
		app_data.load.rect.x       = flagup_width + 2*PADDING + 1;
		app_data.load.rect.y       = PADDING;
		app_data.load.rect.width   = w - 3*PADDING - flagup_width - 1;
		app_data.load.rect.height  = h - 2*PADDING;

		/* Free old and allocate neu chart (load history) pixmap */
		if (app_data.load.chart)
		    XFreePixmap(display, app_data.load.chart);
		w = app_data.load.rect.width * 2;
		h = app_data.load.rect.height;
		app_data.load.chart = XCreatePixmap(display, main_win, w, h, depth);
		XFillRectangle(display, app_data.load.chart, app_data.gc, 0, 0, w, h);

		/* Repaint the application window */
		app_window_paint(&app_data, MF_ALL);
	    }
	    break;

	case ClientMessage:

	    /* Catch ICCCM kill from WM */
	    cm_ptr = (XClientMessageEvent*) &event;
	    if (cm_ptr->message_type == wm_protocols && cm_ptr->data.l[0] == wm_delete_win)
		app_window_close(0);
	    break;

	case DestroyNotify:

	    /* Catch the window "Destroy" event */
	    if (event.xany.window == main_win)
		app_window_close(0);
	    break;
        }
    }
}


/******************************************************************************\
 * get_options()
\******************************************************************************/

int get_options(int argc, char* argv[])
{
    int  i;

    for (i = 1; i < argc; i++)
    {
	const char *opt_str = argv[i];
 
	if (*opt_str++ != '-' || (++i) >= argc)
	{
	    puts(usage_text);
	    return False;
	}

	if (strcmp(opt_str, "display_name") == 0)
	{
	    display_name = argv[i];
	}
	else if (strcmp(opt_str, "geometry") == 0)
	{
	    geom_string = argv[i];
	}
	else if (strcmp(opt_str, "bg") == 0)
	{
	    bg_name = argv[i];
	}
	else if (strcmp(opt_str, "fg") == 0)
	{
	    fg_name = argv[i];
	}
	else if (strcmp(opt_str, "fn") == 0)
	{
	    fn_name = argv[i];
	}
	else if (strcmp(opt_str, "update") == 0)
	{
	    update = atoi(argv[i]);
	}
	else if (strcmp(opt_str, "exec") == 0)
	{
	    exec_name = argv[i];
	}
	else if (strcmp(opt_str, "file") == 0)
	{
	    file_name = argv[i];
	}
	else if (strcmp(opt_str, "wav") == 0)
	{
	    wav_name = argv[i];
	}
	else
	{
	    fputs(usage_text, stderr);
	    return False;
	}
    }

    /* Set some defaults */
    if (bg_name == NULL) bg_name = back_default;
    if (fn_name == NULL) fn_name = font_default;
    if (update  <   1  ) update  = DEF_UPDATE;

    return argc >= 1;
}


/******************************************************************************\
 * mail_check()
\******************************************************************************/

void mail_check(MailStruct* mailbox, Boolean reset)
{
    Boolean readSinceLastWrite = False;
    long    mailboxsize        = 0;

    if (mailbox->check_cmd != NULL)
    {
	waitType wait_status;
	int	 check_status;

#ifdef INTWAITTYPE
	wait_status = system(mailbox->check_cmd);
#else
	wait_status.w_status = system(mailbox->check_cmd);
#endif
	check_status = waitCode(wait_status);

	/* error in sh checkCommand execution */
	if (waitSig(wait_status))
	    check_status = 2;		/* act as if there is no mail */

	switch (check_status)
	{
	case 0: mailboxsize = mailbox->lastsize + 1;
	    break;
	case 2: mailboxsize = 0;
	    break;
	default: /* treat everything else as no change */
	    mailboxsize = mailbox->lastsize;
	    break;
	}
    }
    else
    {
	struct stat st;

	if (stat(mailbox->filename, &st) == 0)
	{
	    mailboxsize = st.st_size;
	    readSinceLastWrite = (st.st_atime > st.st_mtime);
	}
    }

    /*
     * Now check for changes.  If reset is set then we want to pretent that
     * there is no mail.  If the mailbox is empty then we want to turn off
     * the flag.  Otherwise if the mailbox has changed size then we want to
     * put the flag up, unless the mailbox has been read since the last 
     * write.
     *
     * The cases are:
     *    o  forced reset by user                        DOWN
     *    o  no mailbox or empty (zero-sized) mailbox    DOWN
     *    o  if read after most recent write 		 DOWN
     *    o  same size as last time                      no change
     *    o  bigger than last time                       UP
     *    o  smaller than last time but non-zero         UP
     *
     * The last two cases can be expressed as different from last
     * time and non-zero.
     */

    if (reset)
    {
        /* Do forced reset */
	mailbox->flagup = False;
	mailbox->redraw = True;
    }
    else if (mailboxsize == 0)
    {
        /* No mailbox or empty, redraw if this is a change */
	mailbox->flagup = False;
	mailbox->redraw = (mailbox->lastsize > 0) ? True : False;
    }
    else if (readSinceLastWrite)
    { 	
	/* Mailbox has been read after most recent write */
	mailbox->flagup = False;
	mailbox->redraw = True;
    }
    else if (mailboxsize != mailbox->lastsize)
    {
        /* Mail file has different size */
	if (!mailbox->flagup)
	{
	    if (mailbox->sound_cmd)
		system(mailbox->sound_cmd);
	    else
		XBell(display, 100);
	    mailbox->redraw = True;
	    mailbox->flagup = True;
	}
    }
    else
	mailbox->redraw = False;

    mailbox->lastsize = mailboxsize;
}


/******************************************************************************\
 * mail_setup()
 *
 * Get user name for building mailbox
\******************************************************************************/

void mail_setup(MailStruct* mailbox)
{
    char   *username = (char*) getlogin();
    GC     gc        = app_data.clock.gc;
    size_t size;

    memset(mailbox, 0, sizeof(MailStruct));

    if (!username)
    {
	struct passwd *pw = getpwuid(getuid());

	if (!pw)
	{
	    ERROR_MSG(ERR_USERNAME, "(MAILBOX)");
	    app_window_close(-1);
	}
	username = pw->pw_name;
    }

    /* Setup mailbox filename */
    size = strlen(MAILBOX_DIRECTORY) + 1 + strlen(username) + 1;
    mailbox->filename = (char*) save_malloc(size);
    strcpy(mailbox->filename, MAILBOX_DIRECTORY);
    strcat(mailbox->filename, "/");
    strcat(mailbox->filename, username);

    if (wav_name)
    {
	/* Setup sound command */
	size = strlen(wav_name) + 1 + strlen("play ") + 1;
	mailbox->sound_cmd = (char*) save_malloc(size);
	strcpy(mailbox->sound_cmd, "play ");
	strcat(mailbox->sound_cmd, wav_name);
    }

    /* Create flag down/up mailbox icons */
    mailbox->icon_dn = XCreatePixmapFromBitmapData(
	display, main_win, (char*) flagdn_bits, flagdn_width, flagdn_height,
	black_pix, bg_pix, DefaultDepthOfScreen(screen));

    mailbox->icon_up = XCreatePixmapFromBitmapData(
	display, main_win, (char*) flagup_bits, flagup_width, flagup_height,
	black_pix, bg_pix, DefaultDepthOfScreen(screen));

    /* Fill the envelope with white color */
    XSetForeground(display, gc, white_pix);
    XFillRectangle(display, mailbox->icon_up, gc, flagup_x_hot, flagup_y_hot, 3, 4);
    XSetForeground(display, gc, fg_pix);
}


/******************************************************************************\
 * clock_setup()
 *
 * Initialize the clock's data.
 * Geometry is setup via ConfigureNotify event.
\******************************************************************************/

void clock_setup(ClockStruct* clk)
{
    unsigned long  values_mask = GCForeground | GCBackground | GCFillStyle;
    XGCValues	   values;
    struct tm      ts;
    time_t         tv;

    /* Setup font for clock */
    clk->font = XLoadQueryFont(display, fn_name);

    if (clk->font == NULL)
    {
	/* Use system default font */
	GC def_GC = DefaultGC(display, DefaultScreen(display));
	clk->font = XQueryFont(display, XGContextFromGC(def_GC));
	fn_name   = NULL;
    }
    else
    {
	/* Use loaded font */
	values_mask = values_mask | GCFont;
	values.font = clk->font->fid;
    }

    /* Create a graphic context */
    values.foreground = fg_pix;
    values.background = bg_pix;
    values.fill_style = FillSolid;
    clk->gc = XCreateGC(display, main_win, values_mask, &values);

    /* Set initial time string and values */
    time(&tv);
    ts = *localtime(&tv);
    sprintf(clk->text, "%2d:%02d", ts.tm_hour, ts.tm_min);
    clk->minute = ts.tm_min;
    clk->hour   = ts.tm_hour;
}

/******************************************************************************\
 * clock_setup()
 *
 * Initialize the load's data.
 * Geometry is setup via ConfigureNotify event.
\******************************************************************************/

void load_setup(LoadStruct* load)
{
    memset(load, 0, sizeof(LoadStruct));
}


/******************************************************************************\
 * clock_alarm()
 *
 * Handler function for SIGALRM signal.
 * Updates the clock, checks for mail, get actual cpu and mem load. 
\******************************************************************************/

void clock_alarm(int sig_no)
{
    ClockStruct   *clk = &app_data.clock;
    unsigned char draw_flags = MF_LOAD;
    struct tm     ts;
    time_t        tv;

    /* Get time and date */
    time(&tv);
    ts = *localtime(&tv);

    if (ts.tm_min != clk->minute || ts.tm_hour != clk->hour)
    {
	sprintf(clk->text, "%2d:%02d", ts.tm_hour, ts.tm_min);
	clk->minute = ts.tm_min;
	clk->hour   = ts.tm_hour;
	mail_check(&app_data.mail, False);
	draw_flags |= (MF_CLOCK | MF_MAIL);
    }

    app_data.load.xpos++;
    app_window_paint(&app_data, draw_flags);

    /* Register this function as alarm handler */
    if (signal(SIGALRM, clock_alarm) == SIG_ERR)
    {
	ERROR_MSG(ERR_SIGNAL, "(SIGALRM)");
	app_window_close(-1);
    }
    /* Activate alarm and set timeout in seconds */
    alarm(update);
}

/******************************************************************************\
 * app_window_setup()
\******************************************************************************/

void app_window_setup(ClockAppStruct* app)
{
    unsigned long  values_mask = GCForeground | GCBackground | GCFillStyle;
    XGCValues	   values;

    /*
     * For security reasons, we reset our uid/gid after doing the necessary
     * system initialization and before calling any X routines.
     * Reset gid first while still (maybe) root
     */
    InitLoadPoint();
    setgid(getgid());
    setuid(getuid());

    values.foreground = bg_pix;
    values.background = fg_pix;
    values.fill_style = FillSolid;
    app->gc = XCreateGC(display, main_win, values_mask, &values);

    clock_setup(&app->clock);
    mail_setup(&app->mail);
    load_setup(&app->load);
}


/******************************************************************************\
 * app_window_close()
\******************************************************************************/

void app_window_close(int code)
{
    /* Stop alarm handler */
    alarm(0);

    /* Free allocated X-Server resources */
    if (app_data.icon)
	XFreePixmap(display, app_data.icon);
    if (app_data.gc)
	XFreeGC(display, app_data.gc);

    if (app_data.clock.gc)
	XFreeGC(display, app_data.clock.gc);
    if (fn_name && app_data.clock.font)
	XFreeFont(display, app_data.clock.font);

    if (app_data.mail.icon_up)
	XFreePixmap(display, app_data.mail.icon_up);
    if (app_data.mail.icon_dn)
	XFreePixmap(display, app_data.mail.icon_dn);

    if (app_data.load.chart)
	XFreePixmap(display, app_data.load.chart);

    /* Close display and terminate */
    XCloseDisplay(display);
    exit(code);
}


/******************************************************************************\
 * app_window_paint()
\******************************************************************************/

void app_window_paint(ClockAppStruct* app, unsigned mode)
{
    const GC gc = app->clock.gc;

    if (mode & MF_CLEAR)
    {
	XClearWindow(display, main_win);
	draw_border(&app->load.rect, gc, white_pix, black_pix);
    }

    if (mode & MF_CLOCK)
    {
	ClockStruct *clk = &app->clock;
	const int length = strlen(clk->text);
	const int width  = XTextWidth(clk->font, clk->text, length);
	const int y_ofs  = clk->rect.height - (clk->font->ascent + clk->font->descent);
	const int x      = clk->rect.x + (clk->rect.width - width) / 2;
	const int y      = clk->rect.y + clk->font->ascent + y_ofs / 2 + 1;

	XFillRectangles(display, main_win, app->gc, &clk->rect, 1);
	XDrawString(display, main_win, gc, x, y, clk->text, length);
    }

    if (mode & MF_MAIL)
    {
	const MailStruct* mail = &app->mail;
	const Pixmap      icon = (mail->flagup) ? mail->icon_up : mail->icon_dn;
	const short       x    = mail->position.x;
	const short       y    = mail->position.y;

	XCopyArea(display, icon, main_win, gc, 0, 0, flagup_width, flagup_height, x, y);
    }

    if (mode & MF_LOAD)
    {
	LoadStruct*          load   = &app->load;
	const unsigned short width  = load->rect.width;
	const unsigned short height = load->rect.height;
	XRectangle           *rect  = &load->rect;

	if (load->xpos >= width)
	{
	    load->xofs += 2;

	    if (load->xofs >= width)
	    {
		XFillRectangle(display, load->chart, app->gc, 0, 0, (width << 1), height);
		load->xofs = load->xpos = 0;
	    }
	}
	GetLoadPoint(NULL, NULL, &load->load);
	XDrawLine(display, load->chart, gc, load->xpos, height,
		  load->xpos, height-(int)(load->load * height));
	XCopyArea(display, load->chart, main_win, gc, load->xofs, 0, width, height,
		  rect->x, rect->y);
    }
    XFlush(display);
}


/******************************************************************************\
 * draw_border()
\******************************************************************************/

void draw_border(const XRectangle* rect, GC gc, Pixel low, Pixel high)
{
    const short x1 = rect->x - 1;
    const short y1 = rect->y - 1;
    const short x2 = x1 + rect->width  + 2;
    const short y2 = y1 + rect->height + 2;

    /*  +----+
     *  |
     */
    XSetForeground(display, gc, high);
    XDrawLine(display, main_win, gc, x1, y2, x1, y1);
    XDrawLine(display, main_win, gc, x1, y1, x2, y1);

    /*       |
     *  +----+
     */
    XSetForeground(display, gc, low);
    XDrawLine(display, main_win, gc, x2, y1, x2, y2);
    XDrawLine(display, main_win, gc, x2, y2, x1, y2);
    XSetForeground(display, gc, fg_pix);
}

/******************************************************************************/
